/*
 * Copyright © 2016 Daniel Kirchner
 *
 * This file is part of HALMD.
 *
 * HALMD is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this program. If not, see
 * <http://www.gnu.org/licenses/>.
 */

#ifndef HALMD_MDSIM_GPU_POTENTIALS_PAIR_TRUNCATIONS_SMOOTH_R4_HPP
#define HALMD_MDSIM_GPU_POTENTIALS_PAIR_TRUNCATIONS_SMOOTH_R4_HPP

#include <boost/numeric/ublas/io.hpp>
#include <boost/numeric/ublas/matrix.hpp>
#include <cuda_wrapper/cuda_wrapper.hpp>
#include <lua.hpp>
#include <memory>

#include <halmd/io/logger.hpp>
#include <halmd/mdsim/gpu/potentials/pair/truncations/smooth_r4_kernel.hpp>
#include <halmd/utility/matrix_shape.hpp>

namespace halmd {
namespace mdsim {
namespace gpu {
namespace potentials {
namespace pair {
namespace truncations {

/**
 * define smooth potential truncation
 */
template <typename potential_type>
class smooth_r4
  : public potential_type
{
public:
    typedef typename potential_type::float_type float_type;
    typedef typename potential_type::gpu_potential_type parent_potential;
    typedef smooth_r4_kernel::smooth_r4<parent_potential> gpu_potential_type;
    typedef typename potential_type::matrix_type matrix_type;

    template<typename... Args>
    smooth_r4(matrix_type const& cutoff, float_type h, Args&&... args)
            : potential_type (std::forward<Args>(args)...)
            , r_cut_sigma_(check_shape(cutoff, this->sigma()))
            , r_cut_(element_prod(this->sigma(), r_cut_sigma_))
            , rr_cut_(element_prod(r_cut_, r_cut_))
            , en_cut_(this->size1(), this->size2())
            , rri_smooth_(std::pow(h, -2))
            , g_param_(this->size1() * this->size2())
    {

        for (size_t i = 0; i < this->size1(); ++i) {
            for (size_t j = 0; j < this->size2(); ++j) {
                std::tie(std::ignore, en_cut_(i,j)) = potential_type::operator()(rr_cut_(i, j), i, j);
            }
        }

        LOG("potential cutoff length: r_c = " << r_cut_sigma_);
        LOG("potential cutoff energy: U = " << en_cut_);

        cuda::host::vector<float4> param(g_param_.size());
        for (size_t i = 0; i < param.size(); ++i) {
            fixed_vector<float, 3> p;
            p[smooth_r4_kernel::R_CUT] = r_cut_.data()[i];
            p[smooth_r4_kernel::RR_CUT] = rr_cut_.data()[i];
            p[smooth_r4_kernel::EN_CUT] = en_cut_.data()[i];
            param[i] = p;
        }

        cuda::copy(param, g_param_);
    }

    /** bind textures before kernel invocation */
    void bind_textures() const
    {
        cuda::copy(rri_smooth_, smooth_r4_wrapper<parent_potential>::rri_smooth);
        smooth_r4_wrapper<parent_potential>::param.bind(g_param_);
        potential_type::bind_textures();
    }

    matrix_type const& r_cut() const
    {
        return r_cut_;
    }

    float_type r_cut(unsigned a, unsigned b) const
    {
        return r_cut_(a, b);
    }

    float_type rr_cut(unsigned a, unsigned b) const
    {
        return rr_cut_(a, b);
    }

    matrix_type const& r_cut_sigma() const
    {
        return r_cut_sigma_;
    }


    /**
     * Bind class to Lua.
     */
    static void luaopen(lua_State* L)
    {
        using namespace luaponte;
        module(L, "libhalmd")
        [
            namespace_("mdsim")
            [
                namespace_("gpu")
                [
                    namespace_("potentials")
                    [
                        namespace_("pair")
                        [
                            class_<smooth_r4, potential_type, std::shared_ptr<smooth_r4> >()
                                .property("r_cut", (matrix_type const& (smooth_r4::*)() const) &smooth_r4::r_cut)
                                .property("r_cut_sigma", &smooth_r4::r_cut_sigma)

                          , def("smooth_r4", &std::make_shared<smooth_r4
                              , matrix_type const&
                              , float_type
                              , potential_type const&>
                            )
                        ]
                    ]
                ]
            ]
        ];
    }

private:
    /** cutoff length in units of sigma */
    matrix_type r_cut_sigma_;
    /** cutoff length in MD units */
    matrix_type r_cut_;
    /** square of cutoff length */
    matrix_type rr_cut_;
    /** potential energy at cutoff length in MD units */
    matrix_type en_cut_;
    /** smoothing length */
    float_type rri_smooth_;
    /** adapter parameters at CUDA device */
    cuda::vector<float4> g_param_;
};

} // namespace truncations
} // namespace pair
} // namespace potentials
} // namespace gpu
} // namespace mdsim
} // namespace halmd

#endif /* ! HALMD_MDSIM_GPU_POTENTIALS_PAIR_TRUNCATIONS_SMOOTH_R4_HPP */
